Enum. OutOfBoundsError exception
(Elixir v1.18.4)
An exception that is raised when a function expects an enumerable to have a certain size but finds that it is too small.
For example, this is raised by Access.at!/1.