URI. Error exception
(Elixir v1.18.4)
An exception raised when an error occurs when a URI is invalid.
For example, see URI.new!/1.
An exception raised when an error occurs when a URI is invalid.
For example, see URI.new!/1.