Defines a File.Stream struct returned by File.stream!/3.
The following fields are public:
path- the file pathmodes- the file modesraw- a boolean indicating if bin functions should be usedline_or_bytes- if reading should read lines or a given number of bytesnode- the node the file belongs to
Summary
Types
@type t() :: %File.Stream{ line_or_bytes: :line | pos_integer(), modes: [term()], node: node(), path: Path.t(), raw: boolean() }