An exception raised when a system limit has been reached.
For example, this can happen if you try to create an atom that is too large:
iex> String.to_atom(String.duplicate("a", 100_000))
** (SystemLimitError) a system limit has been reached
An exception raised when a system limit has been reached.
For example, this can happen if you try to create an atom that is too large:
iex> String.to_atom(String.duplicate("a", 100_000))
** (SystemLimitError) a system limit has been reached